/////////////////////////////////////////////////////////////////////////////////////////////////
//
//	RB (23/06/2008)
//	Software di acquisizione del Picomperometro versione AH501
//
/////////////////////////////////////////////////////////////////////////////////////////////////

#include "Listener.h"

pthread_mutex_t mutex1 = PTHREAD_MUTEX_INITIALIZER;

//#define DEBUG
 
void* listen ( void *argin)
{
   Listener * listenclass = (Listener *)argin;
   string s,response,data;
   unsigned	read_failures;
   char tmp_buffer[BUFFER_SIZE];
   unsigned char buffer[BUFFER_SIZE];
   int n_read,n_read_tot;
   
   s="";data="";
   read_failures=0;
   
   n_read_tot = 0;
   memset (buffer, 0, sizeof(buffer));
   while (true) {
	   try
		{
			if (listenclass->p_socket == 0)	{
				listenclass->try_open_socket();
			}
			// First of all let's send queued commands if they exists
			AH501_command command = listenclass->get_enqueued_command();
			if (command.cmd_name != "") {
				listenclass->waiting_reply = command.need_reply;
				listenclass->write_socket(command.cmd_name);
			}

			// Now we can read the read buffer
			if ((listenclass->data_binary)) {
				n_read = listenclass->p_socket->recv(tmp_buffer);
				memcpy(buffer+n_read_tot,tmp_buffer,n_read);
				n_read_tot += n_read;
				read_failures=0;
			} else {
				*(listenclass->p_socket) >> response;
				s+=response;
				if (response.length()==0)
					read_failures++;
				else 
					read_failures=0;
			}
		}
	   catch(SocketException & e )
		{
			read_failures++;
		}
	   catch(...)
		{
			read_failures++;
			cout << " ********************** UNKNOWN EXCEPTION **********************" << endl;
			cout << " ********************     listen thread     ********************" << endl;
			cout << " ***************************************************************" << endl;
		}
	   
	   if (read_failures > MAX_READ_FAILURES) {
		read_failures=0;
		//listenclass->restart_socket();
		usleep(3 * USLEEP_DELAY);
	   	continue;
	   }
	   
	   
	   if ((listenclass->data_binary) && !(listenclass->waiting_reply)) {
	   	   // The size of the binary data packet depends on the resolution (res16 => 2Bytes, res24 => 3Bytes)
   		   if (n_read_tot < ( (listenclass->resolution / 8) * NUM_SIGNALS)) {
			usleep(3 * USLEEP_DELAY);
			continue;
			}
	   } else {
		   s = s + '\0';
		   if (s.empty()) {
			usleep(3 * USLEEP_DELAY);
			continue;
			}
	   }
	   
	   pthread_mutex_lock( &mutex1 );
	   if (!(listenclass->data_binary)){
		   while ( s.find("\r\n") != string::npos ) {
	   		//
	   		// Search the reply delimiter "\r\n"
			data.assign(s,0,s.find("\r\n"));
			s.erase(0,s.find("\r\n")+2);
			if (listenclass->waiting_reply) { 
				listenclass->push_reply(data);
			} else {
				listenclass->push_data(data);
			}
		   }
	   } else if ((listenclass->data_binary) && !(listenclass->waiting_reply)) {
		  int remaining = listenclass->push_data(buffer,n_read_tot);
		  // Let's fill the first part of the buffer with remaining chars
		  memcpy(buffer,buffer+n_read_tot-remaining,remaining);
		  n_read_tot = remaining;
	   } else if ((listenclass->data_binary) && (listenclass->waiting_reply)) {
	   	   s = (char *)buffer;
		   memset (buffer, 0, sizeof(buffer));
		   n_read_tot = 0;
		   while ( s.find("\r\n") != string::npos ) {
	   		//
	   		// Search the reply delimiter "\r\n"
			data.assign(s,0,s.find("\r\n"));
			s.erase(0,s.find("\r\n")+2);
			listenclass->push_reply(data);
		        }
	   }
	   pthread_mutex_unlock( &mutex1 );
	   usleep(3 * USLEEP_DELAY);
	   }
   pthread_exit(0);
   return 0;      
}


//+------------------------------------------------------------------
Listener::Listener ( const std::string& host, int port_in)
{
	ip_address=host;
	port = port_in;
	p_socket = 0;
	head = -1;
	circular_flag = false;
	partial_counter = 0;
	waiting_reply = false;
	data_binary = false;
	push_reply(NO_REPLY);
	

	// Launch read_status thread
	//--------------------------------------------
	if (pthread_create( &thread_id, 0, &listen,(void *)this)) {
		cout << "Unable to create listener thread\n";
	} else {
		pthread_detach(thread_id);
	}
	
}



//+------------------------------------------------------------------
Listener::~Listener ()
{
}




//+------------------------------------------------------------------
string Listener::exec_cmd(string command, bool need_reply)
{
	std::string resp;
	short write_ret;
	long stored_head;
	push_reply(NO_REPLY);
	#ifdef DEBUG
	   cout << "exec_cmd ("<< command <<", need_reply = "<< need_reply <<")\n";
	#endif
	enqueue_command(command,need_reply);
	if (need_reply)
		read_reply(resp);
	else
		resp = "ACK";
	return resp;
}

//+------------------------------------------------------------------
//
//	This works only in ASCII mode
//
//+------------------------------------------------------------------

string Listener::get_one_shot_data(long buffer_out[])
{
	std::string resp = "NAK";
	std::string cmd = "G";
	short write_ret;
	long actual_head;
	
	long stored_head = head;
	try
	{
		if (p_socket == 0)
		{
			try_open_socket();
		}
		push_reply(NO_REPLY);
		write_ret = write_socket(cmd);
		long counts = 0;
		while (counts < WAIT_FOR_REPLY){
			actual_head = head;
			if (stored_head != actual_head) {
				break;
				}
			counts++; 
			usleep(USLEEP_DELAY);
			}
		if (counts < WAIT_FOR_REPLY) {
			read_data(buffer_out, 1);
			resp = "ACK";
		}
	}
	catch(SocketException & e )
		{
			cout << "SocketException: Error reading socket " << endl;
		}
	catch(...)
		{
			cout << " ********************** UNKNOWN EXCEPTION **********************" << endl;
			cout << " *******************    get_one_shot_data   ********************" << endl;
			cout << " ***************************************************************" << endl;
		}
	return resp;
}
//***************************************************************************************
//	This method is used for ASCII data format
//***************************************************************************************
void Listener::push_data(string data_in)
{
   unsigned space_counter,space_index;
   string value="",data="";
   long pch, tmp_value;
   string::iterator i;
   short j;
   	 
   // Let's update the circular buffer
   head++;
   partial_counter++;
   if (head == BUFFER_SIZE) {
	head = 0;
	circular_flag = true;
   }

   if (data_binary) {
	#ifdef DEBUG
   	cout << "Data size: " << data_in.size() << endl;
	#endif
   } else {
	   // Let's clean the input string from confusing chars
	   for (i = data_in.begin(); i != data_in.end(); i++) {
		if (((*i >= '0')&&(*i <= '9')) || ((*i >= 'a')&&(*i <= 'z')) || (*i == ' '))
			data.append(1,*i);
	   }
	   // Let's delete the last char if it's a spece
	   if (*(data.end()-1) == ' ')
   		data.erase(data.end()-1);
	   space_counter=0;
	   space_index=0;
	#ifdef DEBUG
	   cout << "push_data ("<< data_in <<") cleared = |"<< data <<"|\n";
	#endif
	   while (data.find(" ",space_index)!= string::npos ) {
   		value = data.substr(space_index,data.find(" ",space_index)-space_index);
		value.insert(0,"0x");
		tmp_value = strtol(value.c_str(),(char **)NULL, 16);
		
		if (resolution == 24) {
			if (tmp_value >= DATA_SIGN_24_BIT) {
				data_buffer[head][space_counter] = (-1) * (DATA_FFFF_24_BIT - tmp_value + 1);
			} else {
				data_buffer[head][space_counter] = tmp_value;
			}
		} else if (resolution == 16) {
			if (tmp_value >= DATA_SIGN_16_BIT) {
				data_buffer[head][space_counter] = (-1) * (DATA_FFFF_16_BIT - tmp_value + 1);
			} else {
				data_buffer[head][space_counter] = tmp_value;
			}
		} else {
			#ifdef DEBUG
			cout << "UNKNOWN RESOLUTION!" << endl;
			#endif
		}

		#ifdef DEBUG
   		cout << data_buffer[head][space_counter] << " ";
		#endif
   		space_index = data.find(" ",space_index)+1;
		space_counter++;
		if (space_counter == NUM_SIGNALS) {
			// More than NUM_SIGNALS in a row, we've missed a '\r\n' ?
			#ifdef DEBUG
			   cout << "TOO MUCH DATA IN A ROW!" << endl;
			#endif
			space_counter = 0;
			head++;
			partial_counter++;

			for (j=0; j <NUM_SIGNALS; j++) {
				if (fabs(data_buffer[head][j]) > MAX_ADC_VALUE) {
					head--;
					partial_counter--;
					#ifdef DEBUG
					   cout << "Data " << j << " overrange: let's ignore last saving" << endl;
					#endif
				}
			}
		}
	   }
	   if (space_counter == (NUM_SIGNALS-1)) {
		value = data.substr(space_index);
		value.insert(0,"0x");
		tmp_value = strtol(value.c_str(),(char **)NULL, 16);
		if (resolution == 24) {
			if (tmp_value >= DATA_SIGN_24_BIT) {
				data_buffer[head][space_counter] = (-1) * (DATA_FFFF_24_BIT - tmp_value + 1);
			} else {
				data_buffer[head][space_counter] = tmp_value;
			}
		} else if (resolution == 16) {
			if (tmp_value >= DATA_SIGN_16_BIT) {
				data_buffer[head][space_counter] = (-1) * (DATA_FFFF_16_BIT - tmp_value + 1);
			} else {
				data_buffer[head][space_counter] = tmp_value;
			}
		} else {
			#ifdef DEBUG
			cout << "UNKNOWN RESOLUTION!" << endl;
			#endif
		}
		#ifdef DEBUG
		cout << data_buffer[head][space_counter] << endl;
		#endif
		for (j=0; j <NUM_SIGNALS; j++) {
			if (fabs(data_buffer[head][j]) > MAX_ADC_VALUE) {
				head--;
				partial_counter--;
				#ifdef DEBUG
				   cout << "Data " << j << " overrange: let's ignore last saving" << endl;
				#endif
				return;
			}
		}

	   } else {
		head--;
		partial_counter--;
		#ifdef DEBUG
		   cout << "Problems reading data: let's ignore last saving" << endl;
		#endif
	   }
   }
}
//***************************************************************************************
//	This method is used for BINARY data format
//
//	Args in: a char array containing data and the datasize 
//	Return value: remaining chars in the array
//***************************************************************************************
int Listener::push_data(unsigned char *buffer,int n_read_tot)
{
  int jj;
  int ii;
  int SingleDataSize = resolution / 8;
   	 
  if (data_binary) {
	jj = 0;
	while ( (n_read_tot - jj) >= (NUM_SIGNALS * SingleDataSize)) {
   		// Let's update the circular buffer
		head++;
		partial_counter++;
		if (head == BUFFER_SIZE) {
			head = 0;
			circular_flag = true;
		}
		for (int kk=0 ; kk < NUM_SIGNALS ; kk++) {
		    switch (SingleDataSize) {
			case 2:
			data_buffer[head][kk] = buffer[jj+(kk *2)]*256 + buffer[jj+(kk *2)+1];
			if (data_buffer[head][kk] >= DATA_SIGN_16_BIT) {
				data_buffer[head][kk] = (-1) * (DATA_FFFF_16_BIT - data_buffer[head][kk] + 1);
			}
			break;
			case 3:
			data_buffer[head][kk] = buffer[jj+(kk *3)] * 65536 + buffer[jj+(kk *3)+1]*256 + buffer[jj+(kk *3)+2];
			if (data_buffer[head][kk] >= DATA_SIGN_24_BIT) {
				data_buffer[head][kk] = (-1) * (DATA_FFFF_24_BIT - data_buffer[head][kk] + 1);
			}
			break;
			default:
			cout << "push_data() ***ERROR Unknown Resolution: " << resolution << endl;
		    }
		}
		jj+=(NUM_SIGNALS*SingleDataSize);
	}
	// Let's fill the first part of the buffer with remaining chars
	ii= n_read_tot - jj;
   } else {
   	   ii=0;
	   cout << "Data Size: " << n_read_tot << endl;
   }
   return ii;
}

//***************************************************************************************

void Listener::push_reply(string data_in)
{
   string::iterator i;
   //cmd_reply.clear();
   cmd_reply.erase(cmd_reply.begin(), cmd_reply.end());
   // Let's clean the input string from confusing chars
   for (i = data_in.begin(); i != data_in.end(); i++) {
	if (((*i >= '0')&&(*i <= '9')) || 
	    ((*i >= 'A')&&(*i <= 'Z')) ||
	    ((*i >= 'a')&&(*i <= 'z')) ||
	     (*i == ' ') || (*i == '.'))
		cmd_reply.append(1,*i);
   }
#ifdef DEBUG
   cout << "push_reply( " << data_in << "), cleared to " << cmd_reply << endl;
#endif
   
}

//***************************************************************************************

void Listener::enqueue_command(string cmd_in, bool need_reply_in)
{
   //string::iterator i;
   //cmd_reply.clear();
   //cmd_queued.erase(cmd_queued.begin(), cmd_queued.end());
   AH501_command AH501_cmd;
   AH501_cmd.cmd_name = cmd_in; 
   AH501_cmd.need_reply = need_reply_in;
   
   cmd_queued.insert(cmd_queued.begin(),AH501_cmd);
   
}

//***************************************************************************************

void Listener::read_data(long buffer_out[],short filter_size)
{
#ifdef DEBUG
   cout << "read_data, with mean filter over " << filter_size << " samples" << endl;
#endif
   
   double Sum[NUM_SIGNALS];
   float wd[NUM_SIGNALS][3];
   int i,j,index;
   for (i=0; i < NUM_SIGNALS; i++)
   	Sum[i] = 0;
   //
   pthread_mutex_lock( &mutex1 );
   //
   index = head;
   j = 0;
   while (j < filter_size) {
   	for (i=0; i < NUM_SIGNALS; i++) {
	   wd[i][j%3] = data_buffer[index][i];
	   //
	   if (j < 3) {
   	      Sum[i] += data_buffer[index][i];
	   } else {
	      // Avoid spikes by using the middle value of the last three values
	      if ((wd[i][0] < wd[i][1] && wd[i][1] < wd[i][2]) || (wd[i][2] < wd[i][1] && wd[i][1] < wd[i][0]))
	         Sum[i] += wd[i][1];
	      else if ((wd[i][1] < wd[i][0] && wd[i][0] < wd[i][2]) || (wd[i][2] < wd[i][0] && wd[i][0] < wd[i][1]))
	         Sum[i] += wd[i][0];
	      else
	         Sum[i] += wd[i][2];
	   }
	}
	j++;
	index --;
	if (index < 0){
		if (circular_flag)
		index = BUFFER_SIZE-1;
	 else
		index = head;
	}
   }
   //
   pthread_mutex_unlock( &mutex1 );
   //
   for (i=0; i < NUM_SIGNALS; i++) {
   	buffer_out[i] = (long)(Sum[i]/filter_size);
   }
}

//***************************************************************************************

void Listener::read_reply(std::string &resp)
{
   int counts=0;
   //
   //
   std::string actual_reply = NO_REPLY;
   while (counts < WAIT_FOR_REPLY){
   	pthread_mutex_lock( &mutex1 );
	actual_reply = cmd_reply;
	pthread_mutex_unlock( &mutex1 );
	if (actual_reply != NO_REPLY) {
		break;
	}
	counts++; 
	usleep(USLEEP_DELAY);
	}
   waiting_reply = false;
   resp = actual_reply;
   //
#ifdef DEBUG
   cout << "read_reply: |" << resp << "| counts= :" << counts << endl;
#endif

}


//***************************************************************************************

void Listener::restart_socket()
{
	string error_message;
#ifdef DEBUG
	cout << "Listener::restart_socket(): entering... !" << endl;
#endif
	if(p_socket)
	{
		delete p_socket;
		p_socket =0;
	}
	try {
		try_open_socket();
	} catch(SocketException & e ) {
		error_message=e.description();
#ifdef DEBUG
		cout << error_message << endl;
#endif
	}
}




//***************************************************************************************

void Listener::try_open_socket()
{
#ifdef DEBUG
	cout << "Listener::try_open_socket(): entering... !" << endl;
#endif
	currentsocketNumberRetries=0;
	string error_message;
	// First close old socket if it exists
	if (p_socket)
	{
		delete p_socket;
		p_socket=0;
		// wait a while for controller to close properly the socket
		sleep(static_cast<unsigned long>(1));
	}
	// We try to open a new socket
	while (currentsocketNumberRetries <= socketNumberRetries)
	{
		try
		{
			p_socket=new ClientSocket(ip_address , port);
			// OK?
			if (p_socket)
			{
				return; // if we get the socket exit
			}
		}
		catch (SocketException & e )
		{
			error_message=e.description();
#ifdef DEBUG
			cout << error_message << endl;
#endif
			delete p_socket;
			p_socket=0;
			currentsocketNumberRetries++;
		}
		catch(...)
		{
			cout << " ********************** UNKNOWN EXCEPTION **********************" << endl;
			cout << " ******************** Listener::try_open_socket ********************" << endl;
			cout << " ***************************************************************" << endl;
		}
		sleep(2);
	} // end while
		// after socketNumberRetries we still cannot get the socket so warn upper layers

	throw SocketException(error_message);

}
	



//***************************************************************************************

short Listener::write_socket(std::string & temp)
{
	std::string::size_type it;
	it = temp.find_last_of(";");
	if(it != std::string::npos)
	{
		temp.erase(it);
	}
	int nb_input_commands= 0;
	if (temp != "S")	temp += "\r";
	try
	{
		// Next lines in order to replace next line
		//nb_input_commands = count(temp.begin(), temp.end(), ';');
		int index=0;
		while (temp.find(';',index) != string::npos) {
			index = temp.find(';',index)+1;
			nb_input_commands ++;
		}
		//
		nb_input_commands ++;
		*p_socket << temp;
	}
	catch(SocketException & e )
	{
		throw(e);
	}
	catch(...)
	{
		cout << " **** UNKNOWN EXCEPTION ***** Listener::write_socket() ****" << endl;
	}
	return(nb_input_commands);
}
	
	
	
//***************************************************************************************

void Listener::read_socket(std::string &s)
{
	std::string response;

	try
	{
		*p_socket >> response;
		s+=response;
	}
	catch(SocketException & e )
	{
		throw(e);
	}
	catch(...)
	{
		cout << " ********************** UNKNOWN EXCEPTION **********************" << endl;
		cout << " ******************** Listener::read_socket ********************" << endl;
		cout << " ***************************************************************" << endl;
	}
	// add a '\0' at end of string
	s = s + '\0';
	return ;
}

//***************************************************************************************

long Listener::get_buffer_head(void)
{
	long val;
	val = head;
	return val;
}

//***************************************************************************************

long Listener::get_buffer_size(void)
{
	return BUFFER_SIZE;
}

//***************************************************************************************

void Listener::reset_partial_counter(void)
{
	partial_counter = 0;
	return;
}

//***************************************************************************************

long Listener::get_partial_counter(void)
{
	long val;
	val = partial_counter;
	return val;
}

//***************************************************************************************

AH501_command Listener::get_enqueued_command(void)
{
	AH501_command AH501_cmd;

	if (cmd_queued.empty()) {
		AH501_cmd.cmd_name = ""; 
		AH501_cmd.need_reply = false;
	} else {
		AH501_cmd = *(cmd_queued.end()-1);
		cmd_queued.pop_back();
	}
	return AH501_cmd;
}
//***************************************************************************************
void Listener::set_resolution(short res_in)
{
	resolution = res_in;
}
