/////////////////////////////////////////////////////////////////////////////////////////////////
//
//	RB (23/06/2008)
//	Software di acquisizione del Picomperometro versione AH501
//
/////////////////////////////////////////////////////////////////////////////////////////////////
#ifndef Listener_class
#define Listener_class

#include <iostream>
#include <pthread.h>
#include <math.h>
#include <string.h>
#include <vector>
#include "ClientSocket.h"
#include "SocketException.h"


#define BUFFER_SIZE	10000
#define NUM_SIGNALS	4
#define BIN_DATA_SIZE	20
#define MAX_ADC_VALUE	16777215
#define REPLY_SIZE	128
#define	socketNumberRetries 3
#define NO_REPLY	"      "
#define NO_CMD	"-"
#define WAIT_FOR_REPLY	1000
#define MAX_READ_FAILURES	500
#define USLEEP_DELAY	1000
#define DATA_FFFF_16_BIT 0xFFFF
#define DATA_SIGN_16_BIT 0x8000
#define DATA_FFFF_24_BIT 0xFFFFFF
#define DATA_SIGN_24_BIT 0x800000

using namespace std;

typedef struct {
   string    cmd_name;
   bool    need_reply;
} AH501_command;

void* listen( void *arg );

class Listener
{
 public:

  Listener ( const std::string& host, int port);
  virtual ~Listener();
  
  string exec_cmd(string command, bool read_reply);
  string get_one_shot_data(long buffer_out[]);
  void reset_partial_counter(void);
  void push_data(string data);
  int push_data(unsigned char *buffer,int n_read_tot);
  void push_reply(string data);
  void enqueue_command(string cmd, bool need_reply);
  void read_data(long data_buffer[],short filter_size);
  void read_reply(std::string &s);
  void restart_socket();
  void try_open_socket();
  short write_socket(std::string & temp);
  void read_socket(std::string &s);
  ClientSocket * p_socket;
  long get_buffer_head(void);
  long get_buffer_size(void);
  long get_partial_counter(void);
  AH501_command get_enqueued_command(void);
  void set_resolution(short);
  
  bool waiting_reply;
  bool data_binary;
  short resolution;

 private:
  long data_buffer[BUFFER_SIZE][NUM_SIGNALS];
  vector<AH501_command> cmd_queued;
  string cmd_reply;
  long head;
  long partial_counter;
  bool circular_flag;
  pthread_t thread_id;
  short currentsocketNumberRetries;
  string ip_address;
  int port;
  
 protected:
     
};


#endif
